import os
import numpy as np
import pandas as pd
import SimpleITK as sitk
from rabies.analysis_pkg.diagnosis_pkg import diagnosis_functions
from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec,
    File, BaseInterface
)


class ScanDiagnosisInputSpec(BaseInterfaceInputSpec):
    CR_dict_file = File(exists=True, mandatory=True, desc="Dictionary with prepared input matrices from confound correction.")
    common_maps_dict_file = File(exists=True, mandatory=True, desc="Brain maps and masks in the commonspace")
    sub_maps_dict_file = File(exists=True, mandatory=True, desc="Brain maps and masks in the subject' space, either native or common.")
    analysis_dict = traits.Dict(
        desc="A dictionary regrouping relevant outputs from analysis.")
    prior_bold_idx = traits.List(
        desc="The index for the ICA components that correspond to bold sources.")
    prior_confound_idx = traits.List(
        desc="The index for the ICA components that correspond to confounding sources.")
    plot_seed_frequencies = traits.Dict(
        desc="Dictionary that pairs a seed name and its index within --seed_list to plot the frequency spectrum")
    figure_format = traits.Str(
        desc="Select file format for figures.")
    nativespace_analysis = traits.Bool(
        desc="Whether input timeseries are in nativespace.")
    native_to_common_transforms = traits.List(
        desc="Transforms to move nativespace computations to commonspace")
    native_to_common_inverses = traits.List(
        desc="Inverses to move nativespace computations to commonspace")
    interpolation = traits.Int(
        desc="Provide an SITK interpolator.")
    brainmap_percent_threshold = traits.Float(
        desc="Input percentage value for thresholding images.")
    rabies_data_type = traits.Int(
        desc="Integer specifying SimpleITK data type.")


class ScanDiagnosisOutputSpec(TraitedSpec):
    figure_temporal_diagnosis = File(
        exists=True, desc="Output figure from the scan diagnosis")
    figure_spatial_diagnosis = File(
        exists=True, desc="Output figure from the scan diagnosis")
    temporal_info = traits.Dict(
        desc="A dictionary regrouping the temporal features.")
    spatial_info = traits.Dict(
        desc="A dictionary regrouping the spatial features.")


class ScanDiagnosis(BaseInterface):
    """
    Extracts several spatial and temporal features on the target scan.
    Spatial features include tSTD, CRsd (variance fitted from confound regression), CR-R^2,
    and network maps from specified BOLD priors at the indices of prior_bold_idx.
    Temporal features include grayplot, 6 motion parameters, framewise displacement,
    DVARS, WM/CSV/edge mask timecourses, CR-R^2, and the average amplitude of BOLD and
    confound components seperately.
    """

    input_spec = ScanDiagnosisInputSpec
    output_spec = ScanDiagnosisOutputSpec

    def _run_interface(self, runtime):
        import pickle
        with open(self.inputs.CR_dict_file, 'rb') as handle:
            CR_data_dict = pickle.load(handle)
        with open(self.inputs.common_maps_dict_file, 'rb') as handle:
            common_maps_data_dict = pickle.load(handle)
        with open(self.inputs.sub_maps_dict_file, 'rb') as handle:
            sub_maps_data_dict = pickle.load(handle)

        figure_format = self.inputs.figure_format

        # convert to an integer list
        prior_bold_idx = [int(i) for i in self.inputs.prior_bold_idx]
        prior_confound_idx = [int(i) for i in self.inputs.prior_confound_idx]

        if self.inputs.nativespace_analysis:
            resampling_specs = {'transforms':self.inputs.native_to_common_transforms,
                                'inverses':self.inputs.native_to_common_inverses,
                                'interpolation':self.inputs.interpolation,
                                'rabies_data_type':self.inputs.rabies_data_type,
                                }
        else:
            resampling_specs = {}

        temporal_info, spatial_info = diagnosis_functions.compute_spatiotemporal_features(
            CR_data_dict, sub_maps_data_dict, common_maps_data_dict, self.inputs.analysis_dict, 
            prior_bold_idx, prior_confound_idx,
            nativespace_analysis=self.inputs.nativespace_analysis,resampling_specs=resampling_specs)

        temporal_fig_list, fig2 = diagnosis_functions.scan_diagnosis(CR_data_dict, common_maps_data_dict, temporal_info,
                                   spatial_info, plot_seed_frequencies=self.inputs.plot_seed_frequencies, 
                                   brainmap_percent_threshold=self.inputs.brainmap_percent_threshold)

        import pathlib
        filename_template = pathlib.Path(CR_data_dict['name_source']).name.rsplit(".nii")[0]
        figure_path = os.path.abspath(filename_template)
        spatial_fig_file = figure_path+f'_spatial_diagnosis.{figure_format}'
        fig2.savefig(spatial_fig_file, bbox_inches='tight')

        if len(temporal_fig_list)==1:
            temporal_fig_file = figure_path+f'_temporal_diagnosis.{figure_format}'
            temporal_fig_list[0].savefig(temporal_fig_file, bbox_inches='tight')
        else: # long scans have figures stacked in PDF
            from matplotlib.backends.backend_pdf import PdfPages
            temporal_fig_file = figure_path+f'_temporal_diagnosis.pdf'
            with PdfPages(temporal_fig_file) as pdf:
                for fig in temporal_fig_list:
                    pdf.savefig(fig, bbox_inches='tight')

        setattr(self, 'figure_temporal_diagnosis',
                temporal_fig_file)
        setattr(self, 'figure_spatial_diagnosis',
                spatial_fig_file)
        setattr(self, 'temporal_info', temporal_info)
        setattr(self, 'spatial_info', spatial_info)

        return runtime

    def _list_outputs(self):
        return {'figure_temporal_diagnosis': getattr(self, 'figure_temporal_diagnosis'),
                'figure_spatial_diagnosis': getattr(self, 'figure_spatial_diagnosis'),
                'temporal_info': getattr(self, 'temporal_info'),
                'spatial_info': getattr(self, 'spatial_info'), }


class DatasetDiagnosisInputSpec(BaseInterfaceInputSpec):
    scan_data_list = traits.List(
        exists=True, mandatory=True, desc="A dictionary regrouping the all required accompanying data per scan.")
    seed_prior_maps = traits.List(
        exists=True, desc="A list of expected network map associated to each seed-FC.")
    outlier_threshold = traits.Float(
        desc="The threshold for modified Z-score to classify outliers.")
    network_weighting = traits.Str(
        desc="Whether network maps are absolute or relative.")
    scan_QC_thresholds = traits.Dict(
        desc="Specifications for scan-level QC thresholds.")
    group_avg_prior = traits.Bool(
        desc="Whether to use the group average (median) as a network prior instead of using an external image.")
    figure_format = traits.Str(
        desc="Select file format for figures.")
    extended_QC = traits.Bool(
        desc="Whether to include image intensity and BOLDsd in the group stats.")
    brainmap_percent_threshold = traits.Float(
        desc="Input percentage value for thresholding images.")
    add_smoothing = traits.Bool(
        desc="When to smooth all brain maps prior to QC metric computation with 0.3mm kernel.")


class DatasetDiagnosisOutputSpec(TraitedSpec):
    analysis_QC = traits.Str(
        exists=True, desc="Output figure from the analysis QC.")


class DatasetDiagnosis(BaseInterface):
    """
    Conducts a group-level correlation analysis to assess artefact effects.
    Computes the voxelwise cross-subject correlation between each spatial features
    from the previously run scan diagnosis.
    """

    input_spec = DatasetDiagnosisInputSpec
    output_spec = DatasetDiagnosisOutputSpec

    def _run_interface(self, runtime):
        import pathlib
        import matplotlib.pyplot as plt
        from rabies.utils import flatten_list
        from .analysis_QC import analysis_QC,QC_distributions

        figure_format = self.inputs.figure_format

        out_dir_global = os.path.abspath('analysis_QC/')
        os.makedirs(out_dir_global, exist_ok=True)
        out_dir_parametric = out_dir_global+'/parametric_stats/'
        os.makedirs(out_dir_parametric, exist_ok=True)
        out_dir_non_parametric = out_dir_global+'/non_parametric_stats/'
        os.makedirs(out_dir_non_parametric, exist_ok=True)
        out_dir_dist = out_dir_global+'/sample_distributions/'
        os.makedirs(out_dir_dist, exist_ok=True)

        merged = flatten_list(list(self.inputs.scan_data_list))
        if len(merged) < 3:
            from nipype import logging
            log = logging.getLogger('nipype.workflow')
            log.warning(
                "Cannot run statistics on a sample size smaller than 3, so dataset diagnosis is not run.")
            setattr(self, 'analysis_QC',
                    out_dir_global)
            return runtime

        template_file = merged[0]['anat_ref_file']
        mask_file = merged[0]['mask_file']
        brain_mask = sitk.GetArrayFromImage(sitk.ReadImage(mask_file))
        volume_indices = brain_mask.astype(bool)

        scan_name_list=[]
        mean_maps=[]
        std_maps=[]
        CRsd_maps=[]
        FD_DVARS_corr_list=[]
        tdof_list=[]
        mean_FD_list=[]
        total_CRsd_list=[]

        FC_maps_dict={}
        FC_maps_dict['DR']=[]
        FC_maps_dict['NPR']=[]
        FC_maps_dict['SBC']=[]
        
        DR_conf_corr_dict={}
        DR_conf_corr_dict['DR']=[]
        DR_conf_corr_dict['NPR']=[]
        DR_conf_corr_dict['SBC']=[]

        for scan_data in merged:
            temporal_std = scan_data['temporal_std']
            CRsd = scan_data['predicted_std']
            if np.median(CRsd)==0:
                # exclude scans with a majority of 0s; indicates misregistration
                continue
            scan_name = pathlib.Path(scan_data['name_source']).name.rsplit(".nii")[0]
            scan_name_list.append(scan_name)
            mean_maps.append(scan_data['voxelwise_mean'])
            std_maps.append(temporal_std)
            CRsd_maps.append(CRsd)
            total_CRsd_list.append(scan_data['CR_global_std'])
            FD_DVARS_corr_list.append(scan_data['FD_DVARS_corr'])
            tdof_list.append(scan_data['tDOF'])
            mean_FD_list.append(scan_data['FD_trace'].to_numpy().mean())

            FC_maps_dict['DR'].append(scan_data['DR_BOLD'])
            if scan_data['NPR_maps'] is not None:
                FC_maps_dict['NPR'].append(scan_data['NPR_maps'])
            if scan_data['seed_map_list'] is not None:
                FC_maps_dict['SBC'].append(scan_data['seed_map_list'])

            # computing the temporal correlation between network and confound timecourses
            DR_confound_time = scan_data['DR_confound_time']
            for network_time,key in zip(
                [scan_data['DR_network_time'],scan_data['NPR_network_time'],scan_data['SBC_network_time']],
                ['DR','NPR','SBC']):
                if network_time is not None:
                    # for each network, compute its confound correlation as mean across all DR confound components
                    corr_list = [np.abs(np.corrcoef(network_time[:,[i]].T,DR_confound_time.T)[0,1:]).mean() for i in range(network_time.shape[1])]
                    DR_conf_corr_dict[key].append(corr_list)
 
        # save the list of the scan names that were included in the group statistics
        pd.DataFrame(scan_name_list).to_csv(f'{out_dir_global}/analysis_QC_scanlist.txt', index=None, header=False)

        from rabies.utils import recover_3D
        std_maps=np.array(std_maps)
        non_zero_voxels = ((std_maps==0).sum(axis=0).astype(bool)==0)
        non_zero_mask = os.path.abspath('non_zero_mask.nii.gz')
        sitk.WriteImage(recover_3D(mask_file, non_zero_voxels.astype(float)), non_zero_mask)

        CRsd_maps=np.array(CRsd_maps)[:,non_zero_voxels]

        corr_variable = []
        variable_name = []
        if self.inputs.extended_QC:
            mean_maps=np.array(mean_maps)[:,non_zero_voxels]
            BOLD_std_maps=np.array(std_maps)[:,non_zero_voxels]
            corr_variable += [mean_maps,BOLD_std_maps]
            variable_name += ['BOLD mean', '$\mathregular{BOLD_{SD}}$']

        corr_variable += [CRsd_maps, np.array(mean_FD_list).reshape(-1,1)]
        variable_name += ['$\mathregular{CR_{SD}}$', 'Mean FD']

        mean_FD_array = np.array(mean_FD_list)
        total_CRsd = np.array(total_CRsd_list)
        FD_DVARS_corr = np.array(FD_DVARS_corr_list)

        # tdof effect; if there's no variability don't compute
        if not np.array(tdof_list).std()==0:
            tdof = np.array(tdof_list).reshape(-1,1)
            corr_variable.append(tdof)
            variable_name.append('tDOF')
            tdof_array = np.array(tdof_list)
        else:
            tdof_array = None

        def change_columns(df):
            columns = list(df.columns)
            i=0
            for column in columns:
                if '$\mathregular{CR_{SD}}$' in column:
                    if 'Overlap:' in column:
                        columns[i] = 'Overlap: Prior - CRsd'
                    if 'Avg.:' in column:
                        columns[i] = 'Avg.: CRsd'
                elif '$\mathregular{BOLD_{SD}}$' in column:
                    if 'Overlap:' in column:
                        columns[i] = 'Overlap: Prior - BOLDsd'
                    if 'Avg.:' in column:
                        columns[i] = 'Avg.: BOLDsd'
                i+=1
            df.columns = columns
            return df
                        

        def analysis_QC_network_i(i,FC_maps,prior_map,non_zero_mask, corr_variable, variable_name, template_file, out_dir_parametric, out_dir_non_parametric,analysis_prefix):

            for non_parametric,out_dir in zip([False, True], [out_dir_parametric, out_dir_non_parametric]):
                with np.errstate(invalid='ignore', divide='ignore'):
                    dataset_stats,fig,fig_unthresholded = analysis_QC(FC_maps, prior_map, non_zero_mask, corr_variable, variable_name, template_file, non_parametric=non_parametric, top_percent=self.inputs.brainmap_percent_threshold, smoothing=self.inputs.add_smoothing)
                df = pd.DataFrame(dataset_stats, index=[1])
                df = change_columns(df)
                df.to_csv(f'{out_dir}/{analysis_prefix}{i}_QC_stats.csv', index=None)
                fig_path = f'{out_dir}/{analysis_prefix}{i}_QC_maps.{figure_format}'
                fig.savefig(fig_path, bbox_inches='tight')
                fig_path = f'{out_dir}/{analysis_prefix}{i}_QC_maps_unthresholded.{figure_format}'
                fig_unthresholded.savefig(fig_path, bbox_inches='tight')

                plt.close(fig)
                plt.close(fig_unthresholded)

        def distribution_network_i(i,prior_map,FC_maps,network_var,DR_conf_corr,FD_DVARS_corr, total_CRsd, mean_FD_array, tdof_array, scan_name_list, outlier_threshold,out_dir_dist,scan_QC_thresholds, analysis_prefix):
            ### PLOT DISTRIBUTIONS FOR OUTLIER DETECTION
            with np.errstate(invalid='ignore', divide='ignore'):
                fig,df,QC_inclusion = QC_distributions(prior_map,FC_maps,network_var,DR_conf_corr,FD_DVARS_corr,total_CRsd, mean_FD_array, tdof_array, scan_name_list, scan_QC_thresholds=scan_QC_thresholds, outlier_threshold=outlier_threshold, top_percent=self.inputs.brainmap_percent_threshold)
            df.to_csv(f'{out_dir_dist}/{analysis_prefix}{i}_outlier_detection.csv', index=None)
            fig_path = f'{out_dir_dist}/{analysis_prefix}{i}_sample_distribution.{figure_format}'
            fig.savefig(fig_path, bbox_inches='tight')
            plt.close(fig)
            return QC_inclusion
        
        def prep_QC_thresholds_i(scan_QC_thresholds, analysis, network_i, num_priors):
            analysis_keys = list(scan_QC_thresholds.keys())
            if not analysis in analysis_keys:
                return {'Dice':None, 'Conf':None, 'Amp':False}
            QC_sub_dict = scan_QC_thresholds[analysis]
            QC_thresholds_i={}
            keys = list(QC_sub_dict.keys())
            for key in ['Dice','Conf']:
                if key in keys:
                    if not type(QC_sub_dict[key]) is list:
                        raise ValueError(f"'{QC_sub_dict[key]}' must be a list.")
                    elif len(QC_sub_dict[key])==0:
                        QC_thresholds_i[key]=None
                    else:
                        if not len(QC_sub_dict[key])==num_priors:
                            raise ValueError(f"The number of Dice thresholds for --scan_QC_thresholds does not match the number of {analysis} networks.")
                        QC_thresholds_i[key]=QC_sub_dict[key][network_i]
                else:
                    QC_thresholds_i[key]=None

            if 'Amp' in keys:
                if not type(QC_sub_dict['Amp']) is bool:
                    raise ValueError(f"'{QC_sub_dict['Amp']}' must be True or False.")
                QC_thresholds_i['Amp']=QC_sub_dict['Amp']
            else:
                QC_thresholds_i['Amp']=False
            return QC_thresholds_i

        scan_QC_thresholds = self.inputs.scan_QC_thresholds


        DR_maps_list=np.array(FC_maps_dict['DR'])

        if self.inputs.group_avg_prior:
            num_priors = DR_maps_list.shape[1]
            prior_maps = np.median(DR_maps_list,axis=0)[:,non_zero_voxels]
        else:
            prior_maps = scan_data['prior_maps'][:,non_zero_voxels]
            num_priors = prior_maps.shape[0]

        for i in range(num_priors):
            if self.inputs.network_weighting=='relative':
                network_var=None
            else:
                # network amplitude as L2-norm of a connectivity map
                network_var = np.sqrt((DR_maps_list[:,i,:] ** 2).sum(axis=1))
            DR_i_scan_QC_thresholds=prep_QC_thresholds_i(scan_QC_thresholds, analysis='DR', network_i=i, num_priors=num_priors)

            FC_maps = DR_maps_list[:,i,non_zero_voxels]
            QC_inclusion = distribution_network_i(i,prior_maps[i,:],FC_maps,network_var,np.array(DR_conf_corr_dict['DR'])[:,i],FD_DVARS_corr,total_CRsd, mean_FD_array, tdof_array, scan_name_list, self.inputs.outlier_threshold, out_dir_dist,scan_QC_thresholds=DR_i_scan_QC_thresholds, analysis_prefix='DR')

            # compute group stats only if there is at least 3 scans
            if QC_inclusion.sum()>2:
                # apply QC inclusion
                FC_maps_ = FC_maps[QC_inclusion,:]
                corr_variable_ = [var[QC_inclusion,:] for var in corr_variable]

                analysis_QC_network_i(i,FC_maps_,prior_maps[i,:],non_zero_mask, corr_variable_, variable_name, template_file, out_dir_parametric, out_dir_non_parametric, analysis_prefix='DR')

        if len(FC_maps_dict['NPR'])>0:
            NPR_maps_list=np.array(FC_maps_dict['NPR'])
            if self.inputs.group_avg_prior:
                num_priors = NPR_maps_list.shape[1]
                prior_maps = np.median(NPR_maps_list,axis=0)[:,non_zero_voxels]
            else:
                prior_maps = scan_data['prior_maps'][:,non_zero_voxels]
                num_priors = prior_maps.shape[0]

            for i in range(num_priors):
                if self.inputs.network_weighting=='relative':
                    network_var=None
                else:
                    # network amplitude as L2-norm of a connectivity map
                    network_var = np.sqrt((NPR_maps_list[:,i,:] ** 2).sum(axis=1))

                NPR_i_scan_QC_thresholds=prep_QC_thresholds_i(scan_QC_thresholds, analysis='NPR', network_i=i, num_priors=num_priors)

                FC_maps = NPR_maps_list[:,i,non_zero_voxels]
                QC_inclusion = distribution_network_i(i,prior_maps[i,:],FC_maps,network_var,np.array(DR_conf_corr_dict['NPR'])[:,i],FD_DVARS_corr,total_CRsd, mean_FD_array, tdof_array, scan_name_list, self.inputs.outlier_threshold, out_dir_dist,scan_QC_thresholds=NPR_i_scan_QC_thresholds, analysis_prefix='NPR')

                # compute group stats only if there is at least 3 scans
                if QC_inclusion.sum()>2:
                    # apply QC inclusion
                    FC_maps_ = FC_maps[QC_inclusion,:]
                    corr_variable_ = [var[QC_inclusion,:] for var in corr_variable]

                    analysis_QC_network_i(i,FC_maps_,prior_maps[i,:],non_zero_mask, corr_variable_, variable_name, template_file, out_dir_parametric, out_dir_non_parametric, analysis_prefix='NPR')

        if len(FC_maps_dict['SBC'])>0:
            seed_maps_list=np.array(FC_maps_dict['SBC'])

            if self.inputs.group_avg_prior:
                num_priors = seed_maps_list.shape[1]
                prior_maps = np.median(seed_maps_list,axis=0)[:,non_zero_voxels]

            # prior maps are provided for seed-FC, tries to run the diagnosis on seeds
            elif len(self.inputs.seed_prior_maps)>0:
                prior_maps=[]
                for prior_map in self.inputs.seed_prior_maps:
                    # resample to match the subject
                    sitk_img = sitk.Resample(sitk.ReadImage(prior_map), sitk.ReadImage(mask_file))
                    prior_maps.append(sitk.GetArrayFromImage(sitk_img)[volume_indices])
                prior_maps = np.array(prior_maps)[:,non_zero_voxels]
                num_priors = prior_maps.shape[0]
            else:
                raise ValueError("Must select either --group_avg_prior or --seed_prior_maps as an option for group statistics with SBC.")
            
            for i in range(num_priors):
                # network amplitude as L2-norm of a connectivity map
                network_var = np.sqrt((seed_maps_list[:,i,:] ** 2).sum(axis=1))

                SBC_i_scan_QC_thresholds=prep_QC_thresholds_i(scan_QC_thresholds, analysis='SBC', network_i=i, num_priors=num_priors)

                FC_maps = seed_maps_list[:,i,non_zero_voxels]
                QC_inclusion = distribution_network_i(i,prior_maps[i,:],FC_maps,network_var,np.array(DR_conf_corr_dict['SBC'])[:,i],FD_DVARS_corr,total_CRsd, mean_FD_array, tdof_array, scan_name_list, self.inputs.outlier_threshold, out_dir_dist,scan_QC_thresholds=SBC_i_scan_QC_thresholds, analysis_prefix='seed_FC')

                # compute group stats only if there is at least 3 scans
                if QC_inclusion.sum()>2:
                    # apply QC inclusion
                    FC_maps_ = FC_maps[QC_inclusion,:]
                    corr_variable_ = [var[QC_inclusion,:] for var in corr_variable]

                    analysis_QC_network_i(i,FC_maps_,prior_maps[i,:],non_zero_mask, corr_variable_, variable_name, template_file, out_dir_parametric, out_dir_non_parametric, analysis_prefix='seed_FC')

        setattr(self, 'analysis_QC',
                out_dir_global)
        return runtime

    def _list_outputs(self):
        return {
            'analysis_QC': getattr(self, 'analysis_QC'),
            }




import os
import pathlib
import SimpleITK as sitk
import numpy as np
from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec,
    File, BaseInterface
)
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from rabies.utils import copyInfo_3DImage, run_command
from simpleitk_timeseries_motion_correction.motion import framewise_register_pair
from simpleitk_timeseries_motion_correction.apply_transforms import framewise_resample_volume

def init_bold_reference_wf(opts, name='gen_bold_ref'):
    # gen_bold_ref_head_start
    """
    The 4D raw EPI file is used to generate a representative volumetric 3D EPI. This volume later becomes the target for 
    motion realignment and the estimation of susceptibility distortions through registration to the structural image. 
    Two iterations of motion realignment to an initial median of the volumes are conducted, then a trimmed mean is 
    computed on the realignment volumes, ignoring 5% extreme, and this average becomes the reference image. The final
    image is then corrected using non-local means denoising (Manjón et al., 2010).

    References:
        Manjón, J. V., Coupé, P., Martí-Bonmatí, L., Collins, D. L., & Robles, M. (2010). Adaptive non-local means 
            denoising of MR images with spatially varying noise levels. Journal of Magnetic Resonance Imaging: 
            JMRI, 31(1), 192–203.

    Command line interface parameters:
        --detect_dummy        Detect and remove initial dummy volumes from the EPI, and generate a reference EPI based on
                            these volumes if detected. Dummy volumes will be removed from the output preprocessed EPI.
                            (default: False)

    Workflow:
        parameters
            opts: command line interface parameters

        inputs
            bold_file: Nifti file with EPI timeseries

        outputs
            ref_image: the reference EPI volume
    """
    # gen_bold_ref_head_end

    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(niu.IdentityInterface(
        fields=['bold_file']), name='inputnode')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['ref_image']),
        name='outputnode')

    n_procs=int(os.environ['RABIES_ITK_NUM_THREADS'])
    gen_ref = pe.Node(EstimateReferenceImage(HMC_level=opts.HMC_level, detect_dummy=opts.detect_dummy, rabies_data_type=opts.data_type),
                      name='gen_ref', mem_gb=2*opts.scale_min_memory, n_procs=n_procs)
    gen_ref.plugin_args = {
        'qsub_args': f'-pe smp {str(2*opts.min_proc)}', 'overwrite': True}

    workflow.connect([
        (inputnode, gen_ref, [
            ('bold_file', 'in_file'),
            ]),
        (gen_ref, outputnode, [('ref_image', 'ref_image')]),
    ])

    return workflow


class EstimateReferenceImageInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="4D EPI file")
    HMC_level = traits.Int(desc="Level for motion correction.")
    detect_dummy = traits.Bool(
        desc="specify if should detect and remove dummy scans, and use these volumes as reference image.")
    rabies_data_type = traits.Int(mandatory=True,
                                  desc="Integer specifying SimpleITK data type.")

class EstimateReferenceImageOutputSpec(TraitedSpec):
    ref_image = File(exists=True, desc="3D reference image")

class EstimateReferenceImage(BaseInterface):
    """
    Given a 4D EPI file, estimates an optimal reference image that could be later
    used for motion estimation and coregistration purposes. If the detect_dummy
    option is selected, it will use detected anat saturated volumes (non-steady
    state). Otherwise, a median of a subset of motion corrected volumes is used.
    In the later case, a first median is extracted from the raw data and used as
    reference for motion correction, then a new median image is extracted from
    the corrected series, and the process is repeated one more time to generate
    a final image reference image.
    The final image is corrected using non-local means denoising. (Manjon et al. Journal of Magnetic Resonance Imaging, June 2010.)
    """

    input_spec = EstimateReferenceImageInputSpec
    output_spec = EstimateReferenceImageOutputSpec

    def _run_interface(self, runtime):

        from nipype import logging
        log = logging.getLogger('nipype.workflow')

        n_procs = int(os.environ['RABIES_ITK_NUM_THREADS']) if "RABIES_ITK_NUM_THREADS" in os.environ else os.cpu_count() # default to number of CPUs

        filename_split = pathlib.Path(self.inputs.in_file).name.rsplit(".nii")
        out_ref_fname = os.path.abspath(
            f'{filename_split[0]}_bold_ref.nii.gz')

        in_nii = sitk.ReadImage(self.inputs.in_file,
                                self.inputs.rabies_data_type)
        if not in_nii.GetDimension()==4:
            raise ValueError(f"Input image {self.inputs.in_file} is not 4-dimensional.")

        dummy_ref = False
        if self.inputs.detect_dummy:
            n_volumes_to_discard = _get_vols_to_discard(in_nii)
            if (not n_volumes_to_discard == 0):
                log.info("Detected "+str(n_volumes_to_discard)
                    + " dummy scans. Taking the mean of these volumes as reference EPI.")
                trimean_array = np.quantile(sitk.GetArrayFromImage(in_nii[:,:,:,:n_volumes_to_discard]), (0.2, 0.5, 0.8), axis=0).mean(axis=0)
                ref_3d = copyInfo_3DImage(sitk.GetImageFromArray(
                    trimean_array, isVector=False), in_nii)
                dummy_ref = True
            else:
                log.info(
                    "Detected no dummy scans. Generating the ref EPI based on multiple volumes.")
                
        if not dummy_ref:
            num_timepoints = in_nii.GetSize()[-1]
            if num_timepoints > 50:
                # select a set of 50 frames spread uniformally across time to avoid temporal biases
                subset_idx = np.linspace(0,num_timepoints-1,50).astype(int)
                subset_img_4d = sitk.JoinSeries([in_nii[:,:,:,int(i)] for i in subset_idx])
            else:
                subset_img_4d = in_nii

            trimean_array = np.quantile(sitk.GetArrayFromImage(subset_img_4d), (0.2, 0.5, 0.8), axis=0).mean(axis=0)
            ref_3d = copyInfo_3DImage(
                sitk.GetImageFromArray(trimean_array, isVector=False), in_nii)

            for round in range(2): # conducted 2 rounds of motion correction and re-calculation of the 3D reference
                transforms = framewise_register_pair(
                    subset_img_4d, 
                    ref_3d, 
                    level=self.inputs.HMC_level, 
                    interpolation=sitk.sitkBSpline5, 
                    max_workers=n_procs)
                
                resampled_img = framewise_resample_volume(
                     subset_img_4d, 
                     ref_3d, 
                     transforms, 
                     interpolation=sitk.sitkBSpline5, 
                     clip_negative=True,
                     extrapolator=False,
                     max_workers=n_procs,
                     )

                trimean_array = np.quantile(sitk.GetArrayFromImage(resampled_img), (0.2, 0.5, 0.8), axis=0).mean(axis=0)
                ref_3d = copyInfo_3DImage(
                    sitk.GetImageFromArray(trimean_array, isVector=False), in_nii)

        # saves the finalized average
        sitk.WriteImage(ref_3d, out_ref_fname)

        # denoise the resulting reference image through non-local mean denoising
        # Denoising reference image.
        command = f'DenoiseImage -d 3 -i {out_ref_fname} -o {out_ref_fname}'
        rc,c_out = run_command(command)

        setattr(self, 'ref_image', out_ref_fname)

        return runtime

    def _list_outputs(self):
        return {'ref_image': getattr(self, 'ref_image')}


def _get_vols_to_discard(img):
    '''
    Takes a nifti file, extracts the mean signal of the first 50 volumes and computes which are outliers.
    is_outlier function: computes Modified Z-Scores (https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm) to determine which volumes are outliers.
    '''
    from nipype.algorithms.confounds import is_outlier
    data_slice = sitk.GetArrayFromImage(img)[:50, :, :, :]
    global_signal = data_slice.mean(axis=-1).mean(axis=-1).mean(axis=-1)
    return is_outlier(global_signal)



import os
import numpy as np
import SimpleITK as sitk
import matplotlib.pyplot as plt
# set a dark background
plt.rcParams.update({
    "lines.color": "white",
    "patch.edgecolor": "white",
    "text.color": "black",
    "axes.facecolor": "white",
    "axes.edgecolor": "lightgray",
    "axes.labelcolor": "white",
    "xtick.color": "white",
    "ytick.color": "white",
    "grid.color": "lightgray",
    "figure.facecolor": "black",
    "figure.edgecolor": "black",
    "savefig.facecolor": "black",
    "savefig.edgecolor": "black"})

"""
Standalone reproduction of Nilearn's 'cold_hot' colormap.

Original reference:
Nilearn developers. Nilearn: Statistical learning for neuroimaging in Python.
URL: https://nilearn.github.io/
"""
from matplotlib.colors import LinearSegmentedColormap
# Segment data extracted from the actual Nilearn 'cold_hot'
_cold_hot_segments = {
    'red': [
        (0.0, 1.0, 1.0),
        (0.126984, 0.0, 0.0),
        (0.5, 0.0, 0.0),
        (0.5, 0.0416, 0.0416),
        (0.682540, 1.0, 1.0),
        (1.0, 1.0, 1.0),
    ],
    'green': [
        (0.0, 1.0, 1.0),
        (0.126984, 1.0, 1.0),
        (0.317461, 0.0, 0.0),
        (0.5, 0.0, 0.0),
        (0.5, 0.0, 0.0),
        (0.682540, 0.0, 0.0),
        (0.873016, 1.0, 1.0),
        (1.0, 1.0, 1.0),
    ],
    'blue': [
        (0.0, 1.0, 1.0),
        (0.317461, 1.0, 1.0),
        (0.5, 0.0416, 0.0416),
        (0.5, 0.0, 0.0),
        (0.873016, 0.0, 0.0),
        (1.0, 1.0, 1.0),
    ]
}
# Create the LinearSegmentedColormap
cold_hot = LinearSegmentedColormap('cold_hot', _cold_hot_segments, N=256)


def otsu_scaling(input_image):
    from skimage.filters import threshold_multiotsu
    
    # the input can be either a nifti file or an SITK image
    if isinstance(input_image, sitk.Image):
        img = input_image
    elif os.path.isfile(input_image):
        img = sitk.ReadImage(input_image)
    else:
        raise ValueError(
            "input_image must be a SimpleITK.Image instance or a path to an existing image file"
        )
    array = sitk.GetArrayFromImage(img)

    thresholds = threshold_multiotsu(array.astype(float).flatten(), classes=5, nbins=100)
    voxel_subset=array[array>thresholds[1]] # clip off the background using otsu thresholds
    # select a maximal value which encompasses 90% of the voxels
    voxel_subset.sort()
    vmax=voxel_subset[int(len(voxel_subset)*0.9)]

    scaled = array/vmax
    scaled_img=sitk.GetImageFromArray(scaled, isVector=False)
    scaled_img.CopyInformation(img)
    return scaled_img


def plot_3d(axes,sitk_img,fig,vmin=0,vmax=1,cmap='gray', alpha=1, cbar=False, threshold=None, planes=('sagittal', 'coronal', 'horizontal'), num_slices=4, slice_spacing=0.1):
    physical_dimensions = (np.array(sitk_img.GetSpacing())*np.array(sitk_img.GetSize()))[::-1] # invert because the array is inverted indices
    array=sitk.GetArrayFromImage(sitk_img)

    array[array==0]=None # set 0 values to be empty

    if not threshold is None:
        array[np.abs(array)<threshold]=None

    slice_0 = (1.0-((num_slices-1)*slice_spacing))/2
    slice_fractions=[slice_0]
    for i in range(1,num_slices):
        slice_fractions.append(slice_0+(i*slice_spacing))

    cbar_list = []
    
    ax_number=0
    if 'sagittal' in planes:
        ax=axes[ax_number]
        ax_number+=1
        empty_slice = np.array([np.nan]).repeat(array.shape[0])[:,np.newaxis]
        slices=empty_slice
        for s in slice_fractions:
            slice=array[::-1,:,int(array.shape[2]*s)]
            slices=np.concatenate((slices,slice,empty_slice),axis=1)
        pos = ax.imshow(slices, extent=[0,physical_dimensions[1]*num_slices,0,physical_dimensions[0]], vmin=vmin, vmax=vmax,cmap=cmap, alpha=alpha, interpolation='none')
        ax.axis('off')
        if cbar:
            cbar_list.append(fig.colorbar(pos, ax=ax))

    if 'coronal' in planes:
        ax=axes[ax_number]
        ax_number+=1
        empty_slice = np.array([np.nan]).repeat(array.shape[0])[:,np.newaxis]
        slices=empty_slice
        for s in slice_fractions:
            slice=array[::-1,int(array.shape[1]*s),:]
            slices=np.concatenate((slices,slice,empty_slice),axis=1)
        pos = ax.imshow(slices, extent=[0,physical_dimensions[2]*num_slices,0,physical_dimensions[0]], vmin=vmin, vmax=vmax,cmap=cmap, alpha=alpha, interpolation='none')
        ax.axis('off')
        if cbar:
            cbar_list.append(fig.colorbar(pos, ax=ax))

    if 'horizontal' in planes:
        ax=axes[ax_number]
        ax_number+=1
        empty_slice = np.array([np.nan]).repeat(array.shape[1])[:,np.newaxis]
        slices=empty_slice
        for s in slice_fractions:
            slice=array[int(array.shape[0]*s),::-1,:]
            slices=np.concatenate((slices,slice,empty_slice),axis=1)
        pos = ax.imshow(slices, extent=[0,physical_dimensions[2]*num_slices,0,physical_dimensions[1]], vmin=vmin, vmax=vmax,cmap=cmap, alpha=alpha, interpolation='none')
        ax.axis('off')
        if cbar:
            cbar_list.append(fig.colorbar(pos, ax=ax))
    return cbar_list



import os
import pathlib
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype.interfaces.utility import Function
from .utils import BIDSDataGraber, convert_to_RAS, correct_oblique_affine, convert_3dWarp, apply_autobox, log_transform_nii
from .commonspace_reg import join_iterables

def init_prep_input_wf(data_dir_path, split_name, scan_info, run_iter, opts, name='prep_input_wf'):
    '''
    This workflow organizes the entire processing.

    **Parameters**

        data_dir_path
            Path to the input data directory with proper BIDS folder structure.
        split_name
            input inherited from the prep_bids_iter function
        scan_info
            input inherited from the prep_bids_iter function
        run_iter
            input inherited from the prep_bids_iter function
        opts
            parser options for preprocess

    **Outputs**

        input_bold_list
            A list of input bold files
        input_anat_list
            A list of input anat files
        prep_bold_list
            A list of bold files formatted for preprocessing
        prep_anat_list
            A list of anat files formatted for preprocessing
        joined_scan_info
            A list of scan_info parameters matching bold files
        joined_run
            A list of run parameters matching bold files
        anat_joined_scan_info
            A list of scan_info parameters matching anat files
    '''

    workflow = pe.Workflow(name=name)

    # set output node
    outputnode = pe.Node(niu.IdentityInterface(
        fields=['input_bold_list', 'input_anat_list', 'prep_bold_list', 'prep_anat_list', 'joined_scan_info', 'joined_run', 'anat_joined_scan_info']),
        name='outputnode')

    # setting up all iterables
    input_main_split = pe.Node(niu.IdentityInterface(fields=['split_name', 'scan_info']),
                         name="input_main_split")
    input_main_split.iterables = [('split_name', split_name),
                            ('scan_info', scan_info)]
    input_main_split.synchronize = True

    bold_selectfiles = pe.Node(BIDSDataGraber(bids_dir=data_dir_path, bids_filter=opts.bids_filter['func']),
                               name='bold_selectfiles')

    # node to conver input image to consistent RAS orientation
    bold_convert_to_RAS_node = pe.Node(Function(input_names=['img_file'],
                                                output_names=['RAS_file'],
                                                function=convert_to_RAS),
                                       name='bold_convert_to_RAS')

    if not opts.oblique2card=='none':
        if opts.oblique2card=='3dWarp':
            correct_oblique=convert_3dWarp
        elif opts.oblique2card=='affine':
            correct_oblique=correct_oblique_affine
        bold_oblique2card_node = pe.Node(Function(input_names=['input'],
                                                    output_names=['output'],
                                                    function=correct_oblique),
                                        name='bold_oblique2card')
        workflow.connect([
            (bold_selectfiles, bold_oblique2card_node, [
                ('out_file', 'input'),
                ]),
            (bold_oblique2card_node, bold_convert_to_RAS_node, [
                ('output', 'img_file'),
                ]),
            ])
    else:
        workflow.connect([
            (bold_selectfiles, bold_convert_to_RAS_node, [
                ('out_file', 'img_file'),
                ]),
            ])

    format_bold_buffer = pe.Node(niu.IdentityInterface(fields=['formatted_bold']),
                                        name="format_bold_buffer")

    if opts.bold_autobox: # apply AFNI's 3dAutobox
        bold_autobox = pe.Node(Function(input_names=['in_file'],
                                                    output_names=['out_file'],
                                                    function=apply_autobox),
                                        name='bold_autobox')
        workflow.connect([
            (bold_convert_to_RAS_node, bold_autobox, [
                ('RAS_file', 'in_file'),
                ]),
            (bold_autobox, format_bold_buffer, [
                ('out_file', 'formatted_bold'),
                ]),
            ])
    else:
        workflow.connect([
            (bold_convert_to_RAS_node, format_bold_buffer, [
                ("RAS_file", "formatted_bold"),
                ]),
            ])

    workflow.connect([
        (input_main_split, bold_selectfiles, [
            ("scan_info", "scan_info"),
            ]),
        ])

    if not opts.bold_only:
        input_run_split = pe.Node(niu.IdentityInterface(fields=['run', 'split_name', 'scan_info']),
                            name="input_run_split")
        input_run_split.itersource = ('input_main_split', 'split_name')
        input_run_split.iterables = [('run', run_iter)]
        input_run_split.synchronize = True

        anat_selectfiles = pe.Node(BIDSDataGraber(bids_dir=data_dir_path, bids_filter=opts.bids_filter['anat']),
                                   name='anat_selectfiles')
        anat_selectfiles.inputs.run = None

        anat_convert_to_RAS_node = pe.Node(Function(input_names=['img_file'],
                                                    output_names=['RAS_file'],
                                                    function=convert_to_RAS),
                                           name='anat_convert_to_RAS')

        if not opts.oblique2card=='none':
            if opts.oblique2card=='3dWarp':
                correct_oblique=convert_3dWarp
            elif opts.oblique2card=='affine':
                correct_oblique=correct_oblique_affine
            anat_oblique2card_node = pe.Node(Function(input_names=['input'],
                                                        output_names=['output'],
                                                        function=correct_oblique),
                                            name='anat_oblique2card')
            workflow.connect([
                (anat_selectfiles, anat_oblique2card_node, [
                    ('out_file', 'input'),
                    ]),
                (anat_oblique2card_node, anat_convert_to_RAS_node, [
                    ('output', 'img_file'),
                    ]),
                ])
        else:
            workflow.connect([
                (anat_selectfiles, anat_convert_to_RAS_node, [
                    ('out_file', 'img_file'),
                    ]),
                ])

        format_anat_buffer = pe.Node(niu.IdentityInterface(fields=['formatted_anat']),
                                            name="format_anat_buffer")


        if opts.log_transform:
            log_anat_node = pe.Node(Function(input_names=['in_nii'],
                                                            output_names=[
                                                                'log_nii'],
                                                            function=log_transform_nii),
                                                    name='log_anat_node')

        if opts.anat_autobox: # apply AFNI's 3dAutobox
            anat_autobox = pe.Node(Function(input_names=['in_file'],
                                                        output_names=['out_file'],
                                                        function=apply_autobox),
                                            name='anat_autobox')
            workflow.connect([
                (anat_convert_to_RAS_node, anat_autobox, [
                    ('RAS_file', 'in_file'),
                    ]),
                ])
            if opts.log_transform:
                workflow.connect([
                    (anat_autobox, log_anat_node, [
                        ("out_file", "in_nii"),
                        ]),
                    (log_anat_node, format_anat_buffer, [
                        ("log_nii", "formatted_anat"),
                        ]),
                    ])
            else:
                workflow.connect([
                    (anat_autobox, format_anat_buffer, [
                        ('out_file', 'formatted_anat'),
                        ]),
                    ])
        else:
            if opts.log_transform:
                workflow.connect([
                    (anat_convert_to_RAS_node, log_anat_node, [
                        ("RAS_file", "in_nii"),
                        ]),
                    (log_anat_node, format_anat_buffer, [
                        ("log_nii", "formatted_anat"),
                        ]),
                    ])
            else:
                workflow.connect([
                    (anat_convert_to_RAS_node, format_anat_buffer, [
                        ("RAS_file", "formatted_anat"),
                        ]),
                    ])


        workflow.connect([
            (input_main_split, input_run_split, [
                ("split_name", "split_name"),
                ("scan_info", "scan_info"),
                ]),
            (input_main_split, anat_selectfiles,
             [("scan_info", "scan_info")]),
            (input_run_split, bold_selectfiles, [
                ("run", "run"),
                ]),
            ])

    if not opts.bold_only:
        workflow, source_join_anat_list, merged_join_anat_list = join_iterables(workflow=workflow, joinsource_list=['input_main_split'], node_prefix='anat_list', num_inputs=3)
        workflow.connect([
            (format_anat_buffer, source_join_anat_list, [
                ("formatted_anat", "file_list0"),
                ]),
            (anat_selectfiles, source_join_anat_list, [
                ("out_file", "file_list1"),
                ]),
            (input_main_split, source_join_anat_list, [
                ("scan_info", "file_list2"),
                ]),
            (merged_join_anat_list, outputnode, [
                ("file_list0", "prep_anat_list"),
                ("file_list1", "input_anat_list"),
                ("file_list2", "anat_joined_scan_info"),
                ]),
            ])
        joinsource_list=['input_run_split','input_main_split']
        bold_join_inputs = 4
    else:
        joinsource_list=['input_main_split']
        bold_join_inputs = 3
    workflow, source_join_bold_list, merged_join_bold_list = join_iterables(workflow=workflow, joinsource_list=joinsource_list, node_prefix='bold_list', num_inputs=bold_join_inputs)

    workflow.connect([
        (format_bold_buffer, source_join_bold_list, [
            ("formatted_bold", "file_list0"),
            ]),
        (bold_selectfiles, source_join_bold_list, [
            ("out_file", "file_list1"),
            ]),
        (merged_join_bold_list, outputnode, [
            ("file_list0", "prep_bold_list"),
            ("file_list1", "input_bold_list"),
            ("file_list2", "joined_scan_info"),
            ]),
        ])
    if opts.bold_only:
        workflow.connect([
            (merged_join_bold_list, outputnode, [
                ("file_list0", "prep_anat_list"),
                ("file_list1", "input_anat_list"),
                ("file_list2", "anat_joined_scan_info"),
                ]),
            (input_main_split, source_join_bold_list, [
                ("scan_info", "file_list2"),
                ]),
            ])
    else:
        workflow.connect([
            (input_run_split, source_join_bold_list, [
                ("scan_info", "file_list2"),
                ("run", "file_list3"),
                ]),
            (merged_join_bold_list, outputnode, [
                ("file_list3", "joined_run"),
                ]),
            ])
 
    return workflow


def match_iterables(scan_info, run, joined_scan_info, joined_run, joined_file_list):
    from rabies.utils import flatten_list
    joined_file_list = flatten_list(joined_file_list)
    joined_scan_info = flatten_list(joined_scan_info)
    if run is not None:
        joined_run = flatten_list(joined_run)
    else: # make run redundant with scan_info so that the code below runs 
        joined_run = joined_scan_info
        run = scan_info

    if not len(joined_file_list)==len(joined_scan_info):
        raise ValueError(f'The file list {joined_file_list} is not the same length as the scan info list {joined_scan_info}.')
    if not len(joined_file_list)==len(joined_run):
        raise ValueError(f'The file list {joined_file_list} is not the same length as the run list {joined_run}.')

    selected_file = None
    for file_i,scan_info_i,run_i in zip(joined_file_list,joined_scan_info, joined_run):
        if (scan_info_i == scan_info) and (run_i == run):
            if selected_file is None:
                selected_file = file_i
            else:
                raise ValueError(f'Duplicate files found for {scan_info} and run {run}.')
    if selected_file is None:
        raise ValueError(f'No file found for {scan_info} and run {run}.')
    return selected_file



from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec,
    File, BaseInterface
)

def prep_bids_iter(layout, bids_filter, bold_only=False, inclusion_list=['all'], exclusion_list=['none']):
    '''
    This function takes as input a BIDSLayout, and generates iteration lists
    for managing the workflow's iterables depending on whether --bold_only is
    selected or not.
    '''
    import pathlib

    scan_info = []
    split_name = []
    structural_scan_list = []
    run_iter = {}
    bold_scan_list = []

    subject_list = layout.get_subject()
    if len(subject_list) == 0:
        raise ValueError(
            "No subject information could be retrieved from the BIDS directory. The 'sub-' specification is mandatory.")

    if not 'subject' in list(bids_filter['func'].keys()):
        # enforce that only files with subject ID are read
        bids_filter['func']['subject']=subject_list

    # create the list for all functional images; this is applying all filters from bids_filter
    bold_bids = layout.get(extension=['nii', 'nii.gz'], **bids_filter['func'])
    if len(bold_bids) == 0:
        raise ValueError(
            f"No functional file were found respecting the functional BIDS spec: {bids_filter['func']}")
    
    # remove subject, session and run; these are used later to target single files
    bids_filter['func'].pop('subject', None)
    bids_filter['func'].pop('session', None)
    bids_filter['func'].pop('run', None)

    # filter inclusion/exclusion lists
    from rabies.utils import filter_scan_inclusion, filter_scan_exclusion
    boldname_list=[pathlib.Path(bold.filename).name.rsplit(".nii")[0] for bold in bold_bids]
    updated_split_name = filter_scan_inclusion(inclusion_list, boldname_list)
    updated_split_name = filter_scan_exclusion(exclusion_list, updated_split_name)
    
    filtered_bold_bids=[]
    for name in updated_split_name:
        for bold in bold_bids:
            if name in bold.filename:
                filtered_bold_bids.append(bold)
    bold_bids = filtered_bold_bids

    bold_dict = {}
    for bold in bold_bids:
        sub = bold.get_entities()['subject']
        try:
            ses = bold.get_entities()['session']
        except:
            ses = None

        try:
            run = bold.get_entities()['run']
        except:
            run = None

        if sub not in list(bold_dict.keys()):
            bold_dict[sub] = {}
        if ses not in list(bold_dict[sub].keys()):
            bold_dict[sub][ses] = {}

        bold_list = layout.get(subject=sub, session=ses, run=run, 
                               extension=['nii', 'nii.gz'], return_type='filename',**bids_filter['func'])
        bold_dict[sub][ses][run] = bold_list

    # if not bold_only, then the bold_list and run_iter will be a dictionary with keys being the anat filename
    # otherwise, it will be a list of bold scans themselves
    for sub in list(bold_dict.keys()):
        for ses in list(bold_dict[sub].keys()):
            if not bold_only:
                anat_list = layout.get(subject=sub, session=ses,
                                       extension=['nii', 'nii.gz'], return_type='filename',**bids_filter['anat'])
                if len(anat_list) == 0:
                    raise ValueError(
                        f'Missing an anatomical image for sub {sub} and ses- {ses}, and the following BIDS specs: {bids_filter["anat"]}')
                if len(anat_list) > 1:
                    raise ValueError(
                        f'Duplicate was found for the anatomical file sub- {sub}, ses- {ses}: {str(anat_list)}')
                file = anat_list[0]
                structural_scan_list.append(file)
                filename_template = pathlib.Path(file).name.rsplit(".nii")[0]
                split_name.append(filename_template)
                scan_info.append({'subject_id': sub, 'session': ses})
                run_iter[filename_template] = []

            for run in list(bold_dict[sub][ses].keys()):
                bold_list = bold_dict[sub][ses][run]
                if len(bold_list) > 1:
                    raise ValueError(
                        f'Duplicate was found for bold files sub- {sub}, ses- {ses} and run {run}: {str(bold_list)}')
                file = bold_list[0]
                bold_scan_list.append(file)
                if bold_only:
                    structural_scan_list.append(file)
                    filename_template = pathlib.Path(
                        file).name.rsplit(".nii")[0]
                    split_name.append(filename_template)
                    scan_info.append(
                        {'subject_id': sub, 'session': ses, 'run': run})
                else:
                    run_iter[filename_template].append(run)

    return split_name, scan_info, run_iter, structural_scan_list, bold_scan_list


class BIDSDataGraberInputSpec(BaseInterfaceInputSpec):
    bids_dir = traits.Str(exists=True, mandatory=True,
                          desc="BIDS data directory")
    bids_filter = traits.Dict(exists=True, mandatory=True,
                         desc="BIDS specs")
    scan_info = traits.Dict(exists=True, mandatory=True,
                            desc="Info required to find the scan")
    run = traits.Any(exists=True, desc="Run number")


class BIDSDataGraberOutputSpec(TraitedSpec):
    out_file = File(
        exists=True, desc="Selected file based on the provided parameters.")


class BIDSDataGraber(BaseInterface):
    """
    This interface will select a single scan from the BIDS directory based on the
    input specifications.
    """

    input_spec = BIDSDataGraberInputSpec
    output_spec = BIDSDataGraberOutputSpec

    def _run_interface(self, runtime):
        if 'run' in (self.inputs.scan_info.keys()):
            run = self.inputs.scan_info['run']
        else:
            run = self.inputs.run

        bids_filter = self.inputs.bids_filter.copy()
        bids_filter['subject'] = self.inputs.scan_info['subject_id']
        bids_filter['session'] = self.inputs.scan_info['session']
        if not run is None:
            bids_filter['run'] = run

        from bids.layout import BIDSLayout
        layout = BIDSLayout(self.inputs.bids_dir, validate=False)
        try:
            file_list = layout.get(extension=['nii', 'nii.gz'], return_type='filename', **bids_filter)
            if len(file_list) > 1:
                raise ValueError(f'Provided BIDS spec lead to duplicates: {bids_filter}')
            elif len(file_list)==0:
                raise ValueError(f'No file for found corresponding to the following BIDS spec: {bids_filter}')
        except:
            raise ValueError(f'Error with BIDS spec: {bids_filter}')

        setattr(self, 'out_file', file_list[0])

        return runtime

    def _list_outputs(self):
        return {'out_file': getattr(self, 'out_file')}


###WRAPPERS FOR AFNI'S FUNCTIONS; NECESSARY TO PREVENT ISSUES WHEN READING INPUTS/OUTPUTS FROM WORKFLOW GRAPH

def apply_despike(in_file):
    import pathlib
    import os
    from rabies.utils import run_command
    split = pathlib.Path(in_file).name.rsplit(".nii")[0]
    out_file = os.path.abspath(f"{split}_despike.nii.gz")
    command = f'3dDespike -prefix {out_file} {in_file}'
    rc,c_out = run_command(command)
    return out_file

def apply_autobox(in_file):
    import pathlib
    import os
    from rabies.utils import run_command
    split = pathlib.Path(in_file).name.rsplit(".nii")[0]
    out_file = os.path.abspath(f"{split}_autobox.nii.gz")
    command = f'3dAutobox -input {in_file} -prefix {out_file} -npad 1'
    rc,c_out = run_command(command)
    return out_file


def convert_to_RAS(img_file, out_dir=None):
    # convert the input image to the RAS orientation convention
    import os
    import nibabel as nb
    img = nb.load(img_file)
    if nb.aff2axcodes(img.affine) == ('R', 'A', 'S'):
        return img_file
    else:
        import pathlib  # Better path manipulation
        split = pathlib.Path(img_file).name.rsplit(".nii")
        if out_dir is None:
            out_file = os.path.abspath(split[0]+'_RAS.nii.gz')
        else:
            out_file = out_dir+'/'+split[0]+'_RAS.nii.gz'
            if not os.path.isdir(out_dir):
                os.makedirs(out_dir)
        nb.as_closest_canonical(img).to_filename(out_file)
        return out_file


def correct_oblique_affine(input):
    from nipype import logging
    log = logging.getLogger('nipype.workflow')

    import numpy as np
    import nibabel as nb
    img = nb.load(input)
    # check if the image is oblique
    if (nb.affines.obliquity(img.affine)!=0).sum()==0:
        # no axis is oblique, thus don't apply correction
        log.info(f"The file {input} is not oblique. No corrections applied.")
        return input
    
    log.info(f"The file {input} is considered oblique. The affine matrix is modified.")
    # nb.affines.obliquity assumes the largest dimension is the one that should correspond to voxel size
    aff = img.affine
    new_aff = np.zeros([4,4])
    vs = nb.affines.voxel_sizes(aff)
    vs_ratio = np.abs(aff[:-1, :-1] / vs)
    for ax in [0,1,2]:
        idx = np.where(vs_ratio[:,ax]==vs_ratio[:,ax].max())[0][0]
        new_aff[idx,ax]=vs[ax]*np.sign(aff[idx,ax])
        
    # copy 4th row/column
    new_aff[3,:] = aff[3,:]
    new_aff[:,3] = aff[:,3]

    import pathlib
    import os
    split = pathlib.Path(input).name.rsplit(".nii")[0]
    output = os.path.abspath(f"{split}_2card.nii.gz")
    nb.Nifti1Image(img.dataobj, new_aff, img.header).to_filename(output)
    return output


def convert_3dWarp(input):
    import pathlib
    import os
    from rabies.utils import run_command
    # apply AFNI's 3dWarp to convert from oblique to cartesian
    split = pathlib.Path(input).name.rsplit(".nii")[0]
    output = os.path.abspath(f"{split}_2card.nii.gz")
    command = f'3dWarp -oblique2card -prefix {output} {input}'
    rc = run_command(command)
    return output


def resample_template(structural_scan_list, bold_scan_list, template_file, mask_file, anatomical_resampling, commonspace_resampling, rabies_data_type, bold_only):
    import os
    import SimpleITK as sitk
    import numpy as np
    from rabies.utils import resample_image_spacing, run_command, flatten_list, get_sitk_header
    from nipype import logging
    log = logging.getLogger('nipype.workflow')

    bold_scan_list = flatten_list(bold_scan_list)
    structural_scan_list = flatten_list(structural_scan_list)

    if anatomical_resampling == 'inputs_defined':
        spacing_list = [get_sitk_header(f).GetSpacing()[:3] for f in structural_scan_list]
        
        # find the lowest dimension across all images and all axes
        low_dim = np.asarray([spacing for spacing in spacing_list]).flatten().min()
        registration_spacing = (low_dim, low_dim, low_dim)
    else:
        shape = anatomical_resampling.split('x')
        registration_spacing = (float(shape[0]), float(shape[1]), float(shape[2]))

    log.info(f"Resampling template to {registration_spacing[0]}x{registration_spacing[1]}x{registration_spacing[2]}mm dimensions for registration steps.")
    registration_template = os.path.abspath("registration_template.nii.gz")
    sitk.WriteImage(resample_image_spacing(sitk.ReadImage(
        template_file, rabies_data_type), registration_spacing), registration_template)

    # also resample the brain mask to ensure stable registrations further down
    registration_mask = os.path.abspath("registration_mask.nii.gz")
    command = f'antsApplyTransforms -d 3 -i {mask_file} -r {registration_template} -o {registration_mask} --verbose -n GenericLabel'
    rc,c_out = run_command(command)

    if commonspace_resampling == 'inputs_defined':
        # if bold_only, bold_scan_list and structural_scan_list are the same; avoid re-loading images if already computed above
        if not (anatomical_resampling == 'inputs_defined' and bold_only):
            spacing_list = [get_sitk_header(f).GetSpacing()[:3] for f in bold_scan_list]
        
        # create a list of all the types of spacing across listed files
        spacing_types = [spacing_list[0]]
        for spacing in spacing_list:
            if not spacing in spacing_types:
                spacing_types.append(spacing)

        if len(spacing_types)>1:
            raise ValueError(f"""The following list of image dimensions have been sampled across the input functional images: {spacing_types}. 
    Because of the conflicting image dimensions, a single resampling resolution cannot be inferred for commonspace. 
    Please manually define the desired commonspace resolution with the --commonspace_resampling parameter to pursue 
    preprocessing.""")
        else:
            commonspace_spacing = spacing_types[0]
        
    else:
        shape = commonspace_resampling.split('x')
        commonspace_spacing = (float(shape[0]), float(shape[1]), float(shape[2]))


    log.info(f"Images will be resampled to a commonspace resolution of {commonspace_spacing[0]}x{commonspace_spacing[1]}x{commonspace_spacing[2]}mm.")
    commonspace_template = os.path.abspath("commonspace_template.nii.gz")
    sitk.WriteImage(resample_image_spacing(sitk.ReadImage(
        template_file, rabies_data_type), commonspace_spacing), commonspace_template)

    return registration_template, registration_mask, commonspace_template


def log_transform_nii(in_nii):
    import pathlib
    import os
    import numpy as np
    import nibabel as nib

    split = pathlib.Path(in_nii).name.rsplit(".nii")[0]
    log_nii = os.path.abspath(f"{split}_log.nii.gz")

    img = nib.load(in_nii)
    img_data = img.get_fdata()
    img_data = np.log10(img_data+1) # adding +1 so that 0 values remain 0 after log transform
    img_data = np.nan_to_num(img_data, posinf=0, neginf=0, nan=0)
    img_data[img_data<0] = 0 # clip all negative values to 0
    nifti_log = nib.Nifti1Image(img_data, affine = img.affine, header = img.header)
    nib.save(nifti_log, log_nii)
    
    return log_nii
